(if (not (file-exists? "ask.rkt"))
    (begin
      (printf "ask.rkt was not found. Please copy your solution to this directory\n")
      (exit -1))
    (void))

(include "io-manager.scm")
(include "ask.rkt")
(require net/url)

(define known-update-date #"29.03.2013")
(define check-latest-enabled #t)


(define tests
  '(("tiny.db" "tiny-test1")
    ("tiny.db" "tiny-test2")
    ("tiny.db" "tiny-test3")
    ("tiny.db" "tiny-test4")
    ("tiny.db" "tiny-test5")
    ("medium.db" "medium-test1")
    ("medium.db" "medium-test2")
    ("medium.db" "medium-test3")
    ("medium.db" "medium-test4")
    ("medium.db" "medium-test5")
    ("medium.db" "medium-test6")
    ("large.db" "large-test1")
    ("large.db" "large-test2")
    ("large.db" "large-test3")
    ("large.db" "large-test4")
    ("large.db" "large-test5")
    ("large.db" "large-test6")
    ("large.db" "large-test7")
    ("large.db" "large-test8")
    ("large.db" "large-test9")
    ("large.db" "bonus-test1")
    ("large.db" "bonus-test2")
    ("large.db" "bonus-test3")
    ("large.db" "bonus-test4")))

(define score-per-test 5)

(define (run-all)
  (let iter ((score 0) (tests tests))
    (if (null? tests)
        (printf "\nSCORE: ~s\n" score)
        (let* ((test (car tests))
               (db-file (complete-name (car test)))
               (test-file (complete-name (cadr test))))
          (printf "Running test. DB: [~s], Test: [~s] .......... " db-file test-file)
          (let* ((outcome (run-test db-file test-file))
                 (delta (if (success? outcome) score-per-test 0)))
            (if (success? outcome)
                (printf "OK\n")
                (begin
                  (printf "FAILED\n")
                  (printf (failure-message outcome))
                  (newline)))
                (iter (+ score delta) (cdr tests)))))))
        
(define (run-test db-file test-file)
  (let* ((input (readInput read (list db-file test-file)))
         (db ((getInputFile input db-file)))
         (test ((getInputFile input test-file)))
         (query (cadar test))
         (expected (cadadr test)))
    (check db query expected)))
      
(define (check db query expected)
  (mbind (try-ask db query)
         (lambda (result)
           (if (check-format result)
               (check-equivalence result expected)
               (failure "Result is not in the expected format")))))
                 
(define (check-equivalence result expected)
  (if (result-match expected result)
      (success null)
      (failure (format "Expected:\n~s\nProvided:\n~s" expected result))))
               
(define (try-ask db query)
  (with-handlers (((lambda (x) #t) (lambda (ex) (failure (exn-message ex)))))
    (success (ask db query))))

(define (check-format result)
  (and (list? result)
       (andmap (lambda (sub)
                 (and (list? sub)
                      (andmap (lambda (binding)
                                (and (list? binding)
                                     (not (null? binding))
                                     (not (null? (cdr binding)))
                                     (null? (cddr binding))
                                     (symbol? (car binding))))
                              sub)))
               result)))
  
(define (mbind result1 f)
  (if (failure? result1)
      result1
      (f (extract-value result1))))

   
(define (result-match provided expected)
  (equivalent-subs? (sort-substitutions provided) (sort-substitutions expected)))

(define (sort-substitutions subs)
  (map (lambda (sub) (sort sub bind-compare)) subs))

(define (equivalent-subs? subs1 subs2)
  (and (= (length subs1) (length subs2))
       (andmap (lambda (sub1) (ormap (lambda (sub2) (equal? sub1 sub2)) subs2)) subs1)))

(define (bind-compare bind1 bind2)
  (string<? (symbol->string (car bind1)) (symbol->string (car bind2))))
                                       

(define (complete-name s) (string-append "tests/" s))


(define (success x) (list #t x))
(define (failure x) (list #f x))
(define success? car)
(define (failure? r) (not (success? r)))
(define extract-value cadr)
(define failure-message cadr)


(run-all)

;; Fetching the homework page and getting the last update date for the tester to see if we're up-to-date.

(define (check-if-last-version)
  (let ((last-version (get-last-modification)))
    (cond
      ((not last-version) (printf "~nCould not obtain latest tester version~n"))
      ((not (equal? last-version known-update-date))
       (printf "~nWARNING! The tester is not up to date. Your version is [~a] while the last version is [~a]~n"
               known-update-date
               last-version)))))
               

(define (get-last-modification)
  (let ((adr "http://elf.cs.pub.ro/pp/teme13/scheme-interogare"))
    (let ((m (regexp-match 
              #px"Versiune tester\\:\\s*([0-9_\\.]+)"
              (get-pure-port (string->url adr)))))
      (if m
          (cadr m)
          #f))))

(if check-latest-enabled
    (check-if-last-version)
    (void))

